package edu.gmu.trimdroid.alloy;


import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

import edu.gmu.trimdroid.AbstractXmlParser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


public class CGToAlloy extends AbstractXmlParser {
	private static final String END_LINE = "\n";
	private static final String TAB = "\t";	
	private List<String> addedNodes = new ArrayList<String>();
	private String outputPath;
	private String appName;
	private String targetApp;
	private StringBuffer formalizedModel = new StringBuffer();

	private static Map<String,CGNode> nodesMap = new HashMap<String,CGNode>();

	public void parseCallGraphXml(String inputXMLPath) {
		Document document = prepareXML(inputXMLPath);

		parseNode(document.getDocumentElement());

		outputAlloy(outputPath, formalizedModel.toString());

	}

	private String getActivityName(String nodeName) {
		String[] tmp = nodeName.split(":")[0].trim().split("\\.");
		String result = tmp[tmp.length - 1];
		if(result.contains("$"))
			result = result.replace("$","____");
		return result;
	}

	private String getHandlerName(String nodeName) {
		String tmp = nodeName.split(":")[1].trim();
		tmp = tmp.split("\\s+")[1];
		tmp = tmp.split("\\(")[0];
		return tmp;
	}	

	public void parseNode(Element rootNode) {
		// do something with the current node instead of System.out
		System.out.println(rootNode.getNodeName());

		String isRoot = "Yes";
		String type = "";

		if(rootNode.getNodeName().equals("cgNode") || rootNode.getNodeName().equals("toNode")) {

			String rootNodeName = ((Element)rootNode).getElementsByTagName("node").item(0).getTextContent();
			
			if(rootNode.getNodeName().equals("toNode")) {
				rootNodeName = ((Element)rootNode).getElementsByTagName("node").item(0).getTextContent();
				rootNodeName = getActivityName(rootNodeName) + "___" +getHandlerName(rootNodeName);
										
				Element parent = (Element)rootNode.getParentNode();
				
				type = parent.getElementsByTagName("type").item(0).getTextContent();
				
				if (type.equals("GUI")) {
					type = parent.getElementsByTagName("label").item(0).getTextContent();
				} 
				
				if(type.equals("N/A")) {
					type = "tokhmi";
				}

				rootNodeName += "___" + type;						

				// if node is a leaf node
//				if(rootNode.getNodeName().equals("node"))
//					if(rootNode.getChildNodes().item(0).getNodeType() != Node.ELEMENT_NODE)
//						isRoot = "No";
				if(rootNodeName.equals("Root"))
					isRoot = "Yes";
				else
					isRoot = "No";					
			}				
			
			if(!addedNodes.contains(rootNodeName)) {
				addedNodes.add(rootNodeName);
				formalize(rootNodeName, rootNode, isRoot);
			}
		 
		}

		NodeList nodeList = rootNode.getChildNodes();		

		if(nodeList != null && nodeList.getLength() > 0) 
			for (int i = 0; i < nodeList.getLength(); i++) {

				Node currentNode = nodeList.item(i);

				if (currentNode.getNodeType() == Node.ELEMENT_NODE) {
					//calls this method for all the children which is Element
					parseNode((Element)currentNode);

				} 

			} 		

	}					

	private void formalize(String rootNodeName, Element rootNode, String isRoot) {
		
		String type = "";

		formalizedModel.append("one sig ").append(rootNodeName).append(" extends node{}{")
		.append(END_LINE).append(TAB);

		NodeList childNodes = ((Element)rootNode).getChildNodes();
		int childCount = 0;
		if(childNodes != null && childNodes.getLength() > 0) 
			for (int j = 0; j < childNodes.getLength(); j++) {
				Node childNode = childNodes.item(j);

				if(childNode.getNodeName().equals("outgoingEdge")) {

					NodeList edgeChildNodes = childNode.getChildNodes();
					if(edgeChildNodes != null && edgeChildNodes.getLength() > 0) 
						for (int p = 0; p < edgeChildNodes.getLength(); p++) {
							Node tempNode =  edgeChildNodes.item(p);

							if(tempNode.getNodeName().equals("toNode")) {

								String childNodeName = ((Element)tempNode).getElementsByTagName("node").item(0).getTextContent();
								type = ((Element)childNode).getElementsByTagName("type").item(0).getTextContent();
								
								if (type.equals("GUI")) {
									type = ((Element)childNode).getElementsByTagName("label").item(0).getTextContent();
								} 
								
								if(type.equals("N/A")) {
									type = "tokhmi";
								}
								
								childNodeName = getActivityName(childNodeName) + "___" +getHandlerName(childNodeName) + "___" + type;

								formalizedModel.append(childNodeName);

								formalizedModel.append(" + ");

								childCount ++;
							}
						}
				}

			}

		if(childCount > 0 ) {
			// remove the last +
			String tmp = formalizedModel.substring(0, formalizedModel.length()-2);
			formalizedModel = new StringBuffer(tmp);

			formalizedModel.append(" = child").append(END_LINE);
		} else
			formalizedModel.append("no child").append(END_LINE);

		formalizedModel.append(TAB).append("isRoot = ").append(isRoot).append(END_LINE);
		formalizedModel.append("}").append(END_LINE).append(END_LINE);

	}
	
	private void outputAlloy(String outputPath, String formalizedModel)	{
		String template =

				"module " + appName + END_LINE + 
				END_LINE + "open behavioralModel"+ END_LINE +
				END_LINE + formalizedModel + END_LINE + END_LINE + 
				"pred findPath{" + END_LINE +
				TAB + "one primePath " + END_LINE +
				TAB + "one path " + END_LINE +
				TAB + "some path.edges" + END_LINE +
				"}" + END_LINE +
				"run findPath";

		try {
			File file = new File(outputPath);

			if (!file.exists()) {
				if (file.mkdirs()) {
					System.out.println("Directory is created!");
				} else {
					System.out.println("Failed to create directory!");
				}
			}

			FileWriter fstream = new FileWriter(outputPath+appName+".als");
			BufferedWriter buffOut = new BufferedWriter(fstream);

			System.out.println("Generating...");
			System.out.println();

			template = template.replaceAll("\\$", "");

			buffOut.append(template);					
			buffOut.close();

		} catch (Exception e) {
			e.printStackTrace();
		}		
	}

	public CGToAlloy (String targetApp, String appName, String inputCallGraph, String outputPath) {
		this.appName = appName;
		this.outputPath = outputPath;
		this.targetApp = targetApp;

		parseCallGraphXml(inputCallGraph);

	}
}
