package edu.gmu.trimdroid.alloy;

import edu.mit.csail.sdg.alloy4.A4Reporter;
import edu.mit.csail.sdg.alloy4.Err;
import edu.mit.csail.sdg.alloy4.ErrorWarning;
import edu.mit.csail.sdg.alloy4compiler.ast.Command;
import edu.mit.csail.sdg.alloy4compiler.ast.ExprVar;
import edu.mit.csail.sdg.alloy4compiler.ast.Module;
import edu.mit.csail.sdg.alloy4compiler.parser.CompUtil;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Options;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Solution;
import edu.mit.csail.sdg.alloy4compiler.translator.TranslateAlloyToKodkod;
import org.apache.commons.io.FileUtils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.XMLEvent;
import javax.xml.transform.stream.StreamSource;

public class Evaluator {

	static final String SOLUTIONS_FOLDER = "/solutions";

	Module root = null;
	A4Solution ans = null;
	static Integer solutionNo = 1;
	static int maxSol = 1000000;

	FileOutputStream oFile;
	PrintStream pPRINT = null;

	public Evaluator(String path) {

        // remove solution folder if already exists
		if(Files.exists(Paths.get(path+SOLUTIONS_FOLDER))) {
			File solutions = new File(path+SOLUTIONS_FOLDER);
            try {
                FileUtils.deleteDirectory(solutions);

            } catch (IOException e) {
                e.printStackTrace();
            }
        }

		File folder = new File(path);
		File[] listOfFiles = folder.listFiles();

		try {
			for (int i = 0; i < listOfFiles.length; i++) {
				if (listOfFiles[i].isFile() && listOfFiles[i].getPath().contains("als") && !listOfFiles[i].getPath().contains("interfaceModel.als") && !listOfFiles[i].getPath().contains("behavioralModel.als")) {
					callAlloyEngine(listOfFiles[i].getPath());
				} 
			}
			//			callAlloyEngine(path);
		} catch (Exception err) {
			err.printStackTrace();
		}

		concatSolutionFiles(path+"solutions");
	}

	// for single File
	public Evaluator(File file) {

		try {
			if(!file.exists()) {
				File folder = new File(file.getParent());
				File[] listOfFiles = folder.listFiles();

				for (int i = 0; i < listOfFiles.length; i++) {
					if (listOfFiles[i].isFile() && listOfFiles[i].getName().contains(file.getName().replace(".als", "")+"_")) {
						callAlloyEngine(listOfFiles[i].getPath());
					} 
				}
			}				
			else
				callAlloyEngine(file.getPath());

			concatSolutionInFolder(file.getParent()+"/solutions/"+file.getName().replace(".als", ""));
			//			callAlloyEngine(path);
		} catch (Exception err) {
			err.printStackTrace();
		}
	}
	// concatenate related solution files to one solution file
	private void concatSolutionFiles(String path) {
		File folder = new File(path);
		File[] solutionFolders = folder.listFiles();
		if(solutionFolders == null)
			return ;
		try {
			for (int i = 0; i < solutionFolders.length; i++) {

				if(solutionFolders[i].isDirectory()) {

					File[] combinationFolders = solutionFolders[i].listFiles();

					int mostItems = 0, index = 0 ;

					for (int j = 0; j < combinationFolders.length; j++) {
						if(combinationFolders[j].isDirectory()) {
							File[] listOfFiles = combinationFolders[j].listFiles();

							if(listOfFiles.length>mostItems) {
								mostItems = listOfFiles.length;
							}
							//							System.out.println(combinationFolders[j].toPath()+">>"+j+">>items>>"+listOfFiles.length);
						}						
					}
					String mergedFilePath = solutionFolders[i].toPath().toString();
					for (int k = 0; k < mostItems; k++) {

						List<File> fileToMerge = new ArrayList<File>();

						for (int j = 0; j < combinationFolders.length; j++) {
							File[] listOfFiles = combinationFolders[j].listFiles();
							//							index = k;

							index = k % (listOfFiles.length) ;

							//							if(k>listOfFiles.length -1) {
							//								index = k % listOfFiles.length;
							//							}

							fileToMerge.add(listOfFiles[index]);
						}	
						mergeXMLs(fileToMerge,mergedFilePath+"___"+k+".xml");						
					}
				}

			}
		} catch (Exception err) {
			err.printStackTrace();
		}
	}

	// concatenate solutions in given Folder
	private void concatSolutionInFolder(String path) {
		File folder = new File(path);
		File[] combinationFolders = folder.listFiles();
		if(combinationFolders == null)
			return;
		try {

			int mostItems = 0, index = 0 ;

			for (int j = 0; j < combinationFolders.length; j++) {
				if(combinationFolders[j].isDirectory()) {
					File[] listOfFiles = combinationFolders[j].listFiles();

					if(listOfFiles.length>mostItems) {
						mostItems = listOfFiles.length;
					}
					//							System.out.println(combinationFolders[j].toPath()+">>"+j+">>items>>"+listOfFiles.length);
				}						
			}

			String mergedFilePath = folder.getParent() + "/" + folder.getName() + "/" + folder.getName();
			for (int k = 0; k < mostItems; k++) {

				List<File> filesToMerge = new ArrayList<File>();

				for (int j = 0; j < combinationFolders.length; j++) {
					File[] listOfFiles = combinationFolders[j].listFiles();
					//							index = k;

					index = k % (listOfFiles.length) ;

					//							if(k>listOfFiles.length -1) {
					//								index = k % listOfFiles.length;
					//							}

					filesToMerge.add(listOfFiles[index]);
				}	
				mergeXMLs(filesToMerge,mergedFilePath+"___"+k+".xml");						
			}
		} catch (Exception err) {
			err.printStackTrace();
		}
	}

	private void mergeXMLs(List<File> targets, String mergedFilePath){

		try {
			Writer outputWriter = new FileWriter(mergedFilePath);
			XMLOutputFactory xmlOutFactory = XMLOutputFactory.newFactory();
			XMLEventWriter xmlEventWriter = xmlOutFactory.createXMLEventWriter(outputWriter);
			XMLEventFactory xmlEventFactory = XMLEventFactory.newFactory();

			xmlEventWriter.add(xmlEventFactory.createStartDocument());
			xmlEventWriter.add(xmlEventFactory.createStartElement("", null, "rootSet"));

			XMLInputFactory xmlInFactory = XMLInputFactory.newFactory();
			for (File rootFile : targets) {
				//				System.out.println("Combining File:" + rootFile.getName());
				XMLEventReader xmlEventReader = xmlInFactory.createXMLEventReader(new StreamSource(rootFile));
				XMLEvent event = xmlEventReader.nextEvent();
				// Skip ahead in the input to the opening document element
				while (event.getEventType() != XMLEvent.START_ELEMENT) {
					event = xmlEventReader.nextEvent();
				}

				do {
					xmlEventWriter.add(event);
					event = xmlEventReader.nextEvent();
				} while (event.getEventType() != XMLEvent.END_DOCUMENT);
				xmlEventReader.close();
			}

			//			System.out.println(">>>>>>" );

			xmlEventWriter.add(xmlEventFactory.createEndElement("", null, "rootSet"));
			xmlEventWriter.add(xmlEventFactory.createEndDocument());

			xmlEventWriter.close();
			outputWriter.close();

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (XMLStreamException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public void callAlloyEngine(String model) throws Err, FileNotFoundException {
		String xmlFile = "";
		String trimmedFilename = model.substring(0, model.length() - 4);

		oFile = new FileOutputStream(trimmedFilename+"_Sols.txt", false);
		pPRINT = new PrintStream(oFile);


		A4Reporter rep = new A4Reporter() {
			@Override
			public void warning(ErrorWarning msg) {
				System.out.print("Relevance Warning:\n" + (msg.toString().trim()) + "\n\n");
				System.out.flush();
			}
		};
		root = CompUtil.parseEverything_fromFile(rep, null, model);

		// Choose some default options for how you want to execute the commands
		A4Options options = new A4Options();
		options.solver = A4Options.SatSolver.SAT4J; //.KK;//.MiniSatJNI; //.MiniSatProverJNI;//.SAT4J;
		options.symmetry = 20;
		options.skolemDepth = 1;

		long start = System.currentTimeMillis();
		SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss");
		sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
		Date startTime = new Date(start);


		for (Command command : root.getAllCommands()) {
			System.out.println("============ Command " + command + ": (" + sdf.format(startTime) + ")============");
			pPRINT.println("============ Command " + command + ": (" + sdf.format(startTime) + ")============");
			try {
				ans = TranslateAlloyToKodkod.execute_command(rep, root.getAllReachableSigs(), command, options);
			} catch (Err err) {
				err.printStackTrace();
			}
			for (ExprVar a : ans.getAllAtoms()) {
				root.addGlobal(a.label, a);
			}
			for (ExprVar a : ans.getAllSkolems()) {
				root.addGlobal(a.label, a);
			}

			String[] tmp = model.split("/");			
			String fileName = tmp[tmp.length-1].trim();
			String fileNamePrefix = fileName.substring(0, fileName.length() - 4);
			
			
			String folderPath ;
			if(fileNamePrefix.contains("___")) {				
				tmp = fileNamePrefix.split("___");
				folderPath = model.replace(fileName, "") + SOLUTIONS_FOLDER +"/" + tmp[0] +"/" + "ComSet" + tmp[1];
			}
			else {
				folderPath = model.replace(fileName, "") + SOLUTIONS_FOLDER +"/" + fileNamePrefix;				
			}

			if (ans.satisfiable()) {
				xmlFile = fileNamePrefix + "_Sol_" + solutionNo + ".xml";

				if (!Files.exists(Paths.get(folderPath))) {
					new File(folderPath).mkdirs();
				}

				ans.writeXML(folderPath+"/"+xmlFile);
			} else {
				noMoreSols();
			}


			while (true) {
				solutionNo = solutionNo + 1;
				if (solutionNo > maxSol) {
					break;
				}
				ans = ans.next();
				if (ans.satisfiable()) {
					System.out.println("---------------------------------");
					System.out.println("Sol: " + solutionNo);
					pPRINT.println("---------------------------------");
					pPRINT.println("Sol: " + solutionNo);
					xmlFile = fileNamePrefix + "_Sol_" + solutionNo + ".xml";

					if (!Files.exists(Paths.get(folderPath))) {
						new File(folderPath).mkdirs();
					}

					ans.writeXML(folderPath+"/"+xmlFile);


				} else {
					noMoreSols();
					break;
				}

			}

		}

	}

	private void noMoreSols() {
		long now = System.currentTimeMillis();
		SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss");
		sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
		Date currentTime = new Date(now);

		System.out.println("No more Satisfying solutions. Finished at: " + sdf.format(currentTime));
		System.out.println("\n-----------------------------------------");
		pPRINT.println("No more Satisfying solutions. Finished at: " + sdf.format(currentTime));
		pPRINT.println("\n-----------------------------------------");
	}


}
