package edu.gmu.trimdroid.alloy;

import edu.gmu.trimdroid.metadata.LayoutMetaData;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

import edu.gmu.trimdroid.AbstractXmlParser;
import edu.gmu.trimdroid.metadata.ui.CheckBox;
import edu.gmu.trimdroid.metadata.ui.CheckBoxPreference;
import edu.gmu.trimdroid.metadata.ui.DatePicker;
import edu.gmu.trimdroid.metadata.ui.EditText;
import edu.gmu.trimdroid.metadata.ui.EditTextPreference;
import edu.gmu.trimdroid.metadata.ui.ListPreference;
import edu.gmu.trimdroid.metadata.ui.NumberPicker;
import edu.gmu.trimdroid.metadata.ui.RadioGroup;
import edu.gmu.trimdroid.metadata.ui.SeekBar;
import edu.gmu.trimdroid.metadata.ui.Spinner;
import edu.gmu.trimdroid.soot.depextract.DependencyExtractor;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


public class LayoutToAlloy extends AbstractXmlParser {
	private static final String END_LINE = "\n";
	private static final String TAB = "\t";	
	private boolean prunning = true;

	private String targetApp;
	private String outputPath;
	private String stringsXml;
	// map widget ID to formalized String
	private Map<String,String> formalizedModelMap = new HashMap<>();
	private LayoutMetaData interfaceModel ;
	private Map<String, HashSet<String>> inputClassesMap;

	private DependencyExtractor dependencyExtractor;

	public void parseLayout(String inputXMLPath, String layoutFileName, boolean considerAtions) {	

		Document document = prepareXML(inputXMLPath);
		StringBuffer formalizedModel = new StringBuffer();

		List<String> actions = listActivityActions(document);
		interfaceModel = new LayoutMetaData(inputXMLPath,stringsXml);

		// EditTexts;
		for(EditText editText:interfaceModel.getEditTexts()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(editText.getId()).append("_TextEdit extends TextEdit{}{").append(END_LINE);
			formalizedModel.append(TAB).append("inputType = ").append(editText.getInputType()).append(END_LINE);
//			Set<String> values = dependencyExtractor.getInputClassesMap().get(editText.getId());
//				int numberOfValues = 0;
//			if (values != null && !values.isEmpty())
//				numberOfValues = values.size();
//			formalizedModel.append(TAB).append("value < ").append(numberOfValues-8).append(END_LINE);
			formalizedModel.append("}").append(END_LINE).append(END_LINE);

			formalizedModelMap.put(editText.getId(), formalizedModel.toString());
		}

		// Radio Buttons
		for(RadioGroup radioGroup:interfaceModel.getRadioGroups()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(radioGroup.getId()).append("_RadioGroup extends RadioGroup{}{").append(END_LINE);
			formalizedModel.append(TAB).append("selected < ").append(radioGroup.getRadioButtonCount()-8).append(END_LINE);
			formalizedModel.append("}").append(END_LINE).append(END_LINE);

			formalizedModelMap.put(radioGroup.getId(), formalizedModel.toString());

		} 

		// DatePicker
		for(DatePicker datePicker:interfaceModel.getDatePickers()) {
			formalizedModel = new StringBuffer();
			formalizedModel.append("one sig ").append(datePicker.getId()).append("_DatePicker extends DatePicker{}{").append(END_LINE);

			formalizedModelMap.put(datePicker.getId(), formalizedModel.toString());
		}

		// NumberPicker
		for(NumberPicker numberPicker:interfaceModel.getNumberPickers()) {
			formalizedModel = new StringBuffer();
			formalizedModel.append("one sig ").append(numberPicker.getTitle()).append("_NumberPicker extends NumberPicker{}{").append(END_LINE);
			formalizedModel.append(TAB).append("inputType = string").append(END_LINE);
			formalizedModel.append("}").append(END_LINE).append(END_LINE);
			formalizedModelMap.put(numberPicker.getId(), formalizedModel.toString());
		}

		// SeekBar
		for(SeekBar seekBar:interfaceModel.getSeekBars()) {
			formalizedModel = new StringBuffer();
			formalizedModel.append("one sig ").append(seekBar.getTitle()).append("_SeekBar extends SeekBar{}").append(END_LINE);

			formalizedModelMap.put(seekBar.getId(), formalizedModel.toString());
		}
		
		// CheckBox
		for(CheckBox checkBox:interfaceModel.getCheckBoxes()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(checkBox.getId()).append("_CheckBox extends CheckBox{}").append(END_LINE);

			formalizedModelMap.put(checkBox.getId(), formalizedModel.toString());
		}

		// Spinner
		for(Spinner spinner:interfaceModel.getSpinners()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(spinner.getId()).append("_Spinner extends Spinner{}{").append(END_LINE);
			formalizedModel.append(TAB).append("selected < ").append(spinner.getSpinnerItemsCount()-8).append(END_LINE);
			formalizedModel.append("}").append(END_LINE).append(END_LINE);

			formalizedModelMap.put(spinner.getId(), formalizedModel.toString());
		} 

		
		// EditTextPrefrences
		for(EditTextPreference editTextPref:interfaceModel.getEditTextPrefs()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(editTextPref.getTitle()).append("_TextEditPref extends TextEditPref{}{").append(END_LINE);
			formalizedModel.append(TAB).append("inputType = ").append(editTextPref.getInputType()).append(END_LINE);
//			Set<String> values = dependencyExtractor.getInputClassesMap().get(editTextPref.getId());
//			int numberOfValues = 0;
//			if (values != null && !values.isEmpty())
//				numberOfValues = values.size();
//			formalizedModel.append(TAB).append("value < ").append(numberOfValues-8).append(END_LINE);
			formalizedModel.append("}").append(END_LINE).append(END_LINE);

			formalizedModelMap.put(editTextPref.getId(), formalizedModel.toString());
		}

		// ListPrefrences
		for(ListPreference listPref:interfaceModel.getListPrefs()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(listPref.getTitle()).append("_ListPref extends ListPref{}{").append(END_LINE);
			formalizedModel.append(TAB).append("selected < ").append(listPref.getListItemsCount()-8).append(END_LINE);
			formalizedModel.append("}").append(END_LINE).append(END_LINE);

			formalizedModelMap.put(listPref.getId(), formalizedModel.toString());
		}
		
		// CheckBoxPrefrences
		for(CheckBoxPreference checkBoxPref:interfaceModel.getCheckBoxePrefs()) {
			formalizedModel = new StringBuffer();

			formalizedModel.append("one sig ").append(checkBoxPref.getTitle()).append("_CheckBoxPref extends CheckBoxPref{}").append(END_LINE);

			formalizedModelMap.put(checkBoxPref.getId(), formalizedModel.toString());
		}

		if(considerAtions)
			applyPrunning(layoutFileName, actions);
		else
			applyPrunning(layoutFileName);

	}

	private void applyPrunning(String fileName) {
		// initialize remainingWidgets to all widgets on the screen
		List<String> remainingWidgets = interfaceModel.getWidgetsIdList();
		int counter = 0;

		if(prunning) {
			List<List<String>> dependecies = dependencyExtractor.getConditionalDependecies();

			for(List<String> dependencyGroup:dependecies) {
				generateAlloyForWidgets(fileName+"_"+counter, dependencyGroup);
				counter ++;
				
				// remove the widget from remaining widgets
				for(String s:dependencyGroup) {
					if(remainingWidgets.contains(s))
						remainingWidgets.remove(s);			
				}
				
			}
		}
		// generate alloy for all widgets
		else {
			generateAllCombinations(fileName);
			// clear remainingwidgets
			remainingWidgets = new ArrayList<>();
		}

		// generate alloy for the remaining widgets one by one
		for(String widgetId:remainingWidgets) {
			generateAlloyForSingleWidget(fileName+"___"+counter, widgetId);
			counter ++;
		}
	}

	private void applyPrunning(String fileName, List<String> actions)
	{	

		// generate an alloy for each action
		for(String actionId:actions) {

			// create a unique number for each alloy file
			int counter = 0;
			
			List<String> handlerDependencies = dependencyExtractor.findHandlerDependencies(actionId);
			// initialize remainingWidgets to all widgets on the screen
			List<String> remainingWidgets = interfaceModel.getWidgetsIdList();

			if(this.prunning) {
				if(!handlerDependencies.isEmpty() && handlerDependencies!=null) {
					// generate alloy for only condition dependent widgets dependent to handler
					List<List<String>> dependecies = dependencyExtractor.findConditionalDependencies4Handler(handlerDependencies);						

					for(List<String> dependencyGroup:dependecies) {
						generateAlloyForWidgets(fileName+"$$$"+actionId+"___"+counter, dependencyGroup);
						counter ++;
						// remove the widget from remaining widgets
						for(String s:dependencyGroup) {
							if(remainingWidgets.contains(s))
								remainingWidgets.remove(s);
							// remove textedits from remaining widgets to avoid duplicates
							// TODO: fix this for symbex
//							if (interfaceModel.getIdToType(s).equals(EditText.TYPE)) 
//								remainingWidgets.remove(s);					
						}
						
					}

					/*	for(String s:dependecies) {
							if(!targetWidgets.contains(s))
								targetWidgets.add(s);
						}
					 */

					// generate alloy for the remaining widgets one by one
					for(String widgetId:remainingWidgets) {
						generateAlloyForSingleWidget(fileName+"$$$"+actionId+"___"+counter, widgetId);
						counter ++;
					}
				} 
				else {
					// generate empty alloy just with action and textBoxes
					generateAlloyForSingleWidget(fileName+"$$$"+actionId+"___"+counter, "EMPTY_FILE");
					counter ++;
				}
			}
			else {

				// generate alloy for all widgets
				generateAllCombinations(fileName+"$$$"+actionId+"___"+counter);
				counter ++;
				// clear remainingwidgets
				remainingWidgets = new ArrayList<>();

				// generate alloy for the remaining widgets one by one
				for(String widgetId:remainingWidgets) {
					generateAlloyForSingleWidget(fileName+"$$$"+actionId+"___"+counter, widgetId);
					counter ++;
				}
			}
		} 
	}	

	private void generateAlloyForSingleWidget(String fileName, String targetWidget) {
		String seekBarCountText;
		String numberPickerCountText;
		String checkBoxCountText;
		String textEditCountText;
		String spinnerCountText;
		String datePickerCountText;
		String radioGroupCountText;
		String checkBoxPrefCountText;
		String textEditPrefCountText;
		String listPrefCountText;
		String moduleName = fileName.replace("$$$", "_");
		Map<String,String> formalizedModelMap = new HashMap<>();
		moduleName = moduleName.replace(".xml", "");
		moduleName = moduleName.replace("#", "_");
		int seekBarCount = 0, numberPickerCount = 0, checkBoxCount = 0, textEditCount = 0 , checkBoxPrefCount = 0, listPrefCount = 0, textEditPrefCount = 0 , spinnerCount = 0 , datePickerCount = 0 , radioGroupCount = 0;

		StringBuffer formalizedModel = new StringBuffer();

		if(!targetWidget.equals("EMPTY_FILE")) {

			formalizedModel.append(this.formalizedModelMap.get(targetWidget));
			String type = interfaceModel.getIdToType(targetWidget);

			if(type.equals(EditText.TYPE)) 
				textEditCount ++;
			if(type.equals(RadioGroup.TYPE)) 
				radioGroupCount ++;
			if(type.equals(DatePicker.TYPE)) 
				datePickerCount ++;
			if(type.equals(Spinner.TYPE)) 
				spinnerCount ++;
			if(type.equals(CheckBox.TYPE)) 
				checkBoxCount ++;
			if(type.equals(ListPreference.TYPE)) 
				listPrefCount ++;
			if(type.equals(EditTextPreference.TYPE)) 
				textEditPrefCount ++;
			if(type.equals(CheckBoxPreference.TYPE)) 
				checkBoxPrefCount ++;
			if(type.equals(SeekBar.TYPE)) 
				seekBarCount ++;
			if(type.equals(NumberPicker.TYPE)) 
				numberPickerCount ++;
		}
		// add text boxes and one option for each widget
		else {
			List<String> widgets = interfaceModel.getWidgetsIdList();
			for(String w:widgets) {
				String type = interfaceModel.getIdToType(w);
				String model = this.formalizedModelMap.get(w);
				StringBuffer sb;

				if(type.equals(EditText.TYPE)) {
					formalizedModelMap.put(w, model);
					textEditCount ++;
				}
				if(type.equals(RadioGroup.TYPE)) {
					model = model.split("\\{")[0];
					sb = new StringBuffer(model);
					sb.append("{}{").append(END_LINE).append("selected < -7").append(END_LINE).append("}").append(END_LINE);
					formalizedModelMap.put(w, sb.toString());
					radioGroupCount ++;
				}
				if(type.equals(DatePicker.TYPE)) 
					continue;
				if(type.equals(Spinner.TYPE) || type.equals(ListPreference.TYPE)) {
					model = model.split("\\{")[0];
					sb = new StringBuffer(model);
					sb.append("{}{").append(END_LINE).append("selected < -7").append(END_LINE).append("}");
					formalizedModelMap.put(w, sb.toString());
					spinnerCount ++;
				}
				if(type.equals(CheckBox.TYPE)) 
					continue;
				if(type.equals(EditTextPreference.TYPE)) 
					textEditPrefCount ++;
				if(type.equals(CheckBoxPreference.TYPE)) 
					continue;

				formalizedModel.append(formalizedModelMap.get(w));
			}
		}

		if(checkBoxCount ==0 ) 
			checkBoxCountText = "no CheckBox";
		else
			checkBoxCountText = "#CheckBox = " + checkBoxCount;

		if(textEditCount ==0 ) 
			textEditCountText = "no TextEdit";
		else
			textEditCountText = "#TextEdit = " + textEditCount;

		if(checkBoxPrefCount ==0 ) 
			checkBoxPrefCountText = "no CheckBoxPref";
		else
			checkBoxPrefCountText = "#CheckBoxPref = " + checkBoxPrefCount;

		if(textEditPrefCount ==0 ) 
			textEditPrefCountText = "no TextEditPref";
		else
			textEditPrefCountText = "#TextEditPref = " + textEditPrefCount;

		if(spinnerCount ==0 ) 
			spinnerCountText = "no Spinner";
		else
			spinnerCountText = "#Spinner = " + spinnerCount;

		if(datePickerCount ==0 ) 
			datePickerCountText = "no DatePicker";
		else
			datePickerCountText = "#DatePicker = " + datePickerCount;

		if(radioGroupCount ==0 ) 
			radioGroupCountText = "no RadioGroup";
		else
			radioGroupCountText = "#RadioGroup = " + radioGroupCount;

		if(radioGroupCount ==0 ) 
			radioGroupCountText = "no RadioGroup";
		else
			radioGroupCountText = "#RadioGroup = " + radioGroupCount;

		if(listPrefCount ==0 ) 
			listPrefCountText = "no ListPref";
		else
			listPrefCountText = "#ListPref = " + listPrefCount;
		
		if(seekBarCount ==0 ) 
			seekBarCountText = "no SeekBar";
		else
			seekBarCountText = "#SeekBar = " + seekBarCount;

		if(numberPickerCount ==0 ) 
			numberPickerCountText = "no NumberPicker";
		else
			numberPickerCountText = "#NumberPicker = " + numberPickerCount;

		String template =
				"module " + moduleName + END_LINE + 
				END_LINE + "open interfaceModel"+ END_LINE +
				END_LINE + formalizedModel.toString() + END_LINE + END_LINE + 
				"pred show{" + END_LINE +
				TAB + checkBoxCountText + END_LINE +
				TAB + textEditCountText + END_LINE +
				TAB + spinnerCountText + END_LINE +
				TAB + datePickerCountText + END_LINE +
				TAB + radioGroupCountText + END_LINE +
				TAB + textEditPrefCountText + END_LINE +
				TAB + checkBoxPrefCountText + END_LINE +
				TAB + listPrefCountText + END_LINE +
				TAB + numberPickerCountText + END_LINE +
				TAB + seekBarCountText + END_LINE +
				"}" + END_LINE +
				"run show";

		if (fileName.contains(".xml"))
			fileName = fileName.replace(".xml", "");

		writeToFile(template, fileName);
	}

	private void generateAlloyForWidgets(String fileName, List<String> targetWidgets) {
		String seekBarCountText;
		String numberPickerCountText;
		String checkBoxCountText;
		String textEditCountText;
		String spinnerCountText;
		String datePickerCountText;
		String radioGroupCountText;
		String checkBoxPrefCountText;
		String textEditPrefCountText;
		String listPrefCountText;
		String moduleName = fileName.replace("$$$", "_");
		moduleName = moduleName.replace(".xml", "_");
		moduleName = moduleName.replace("#", "_");
		int seekBarCount = 0, numberPickerCount = 0, checkBoxCount = 0, textEditCount = 0 , checkBoxPrefCount = 0, listPrefCount = 0, textEditPrefCount = 0 , spinnerCount = 0 , datePickerCount = 0 , radioGroupCount = 0;

		StringBuffer formalizedModel = new StringBuffer();
		targetWidgets.removeAll(Collections.singleton(null));

		List<String> producedEditTexts = new ArrayList<String>();

		for (String widgetID : targetWidgets) {

			formalizedModel.append(formalizedModelMap.get(widgetID));
			String type = interfaceModel.getIdToType(widgetID);
			
			if (type == null)
				return;
			
			if(type.equals(EditText.TYPE)) {
				textEditCount ++;
				producedEditTexts.add(widgetID);
			}
			if(type.equals(RadioGroup.TYPE)) 
				radioGroupCount ++;
			if(type.equals(DatePicker.TYPE)) 
				datePickerCount ++;
			if(type.equals(Spinner.TYPE)) 
				spinnerCount ++;
			if(type.equals(CheckBox.TYPE)) 
				checkBoxCount ++;
			if(type.equals(EditTextPreference.TYPE)) 
				textEditPrefCount ++;
			if(type.equals(CheckBoxPreference.TYPE)) 
				checkBoxPrefCount ++;
			if(type.equals(ListPreference.TYPE)) 
				listPrefCount ++;
			if(type.equals(SeekBar.TYPE)) 
				seekBarCount ++;
			if(type.equals(NumberPicker.TYPE)) 
				numberPickerCount ++;
		}
		List<EditText> editTexts = interfaceModel.getEditTexts();

//		for (EditText editText:editTexts) {
//			if(!producedEditTexts.contains(editText.getId())) {
//				textEditCount ++;	
//				formalizedModel.append(formalizedModelMap.get(editText.getId()));
//			}
//		}

		if(checkBoxCount ==0 ) 
			checkBoxCountText = "no CheckBox";
		else
			checkBoxCountText = "#CheckBox = " + checkBoxCount;

		if(textEditCount ==0 ) 
			textEditCountText = "no TextEdit";
		else
			textEditCountText = "#TextEdit = " + textEditCount;

		if(checkBoxPrefCount ==0 ) 
			checkBoxPrefCountText = "no CheckBoxPref";
		else
			checkBoxPrefCountText = "#CheckBoxPref = " + checkBoxPrefCount;

		if(textEditPrefCount ==0 ) 
			textEditPrefCountText = "no TextEditPref";
		else
			textEditPrefCountText = "#TextEditPref = " + textEditPrefCount;

		if(spinnerCount ==0 ) 
			spinnerCountText = "no Spinner";
		else
			spinnerCountText = "#Spinner = " + spinnerCount;

		if(datePickerCount ==0 ) 
			datePickerCountText = "no DatePicker";
		else
			datePickerCountText = "#DatePicker = " + datePickerCount;

		if(radioGroupCount ==0 ) 
			radioGroupCountText = "no RadioGroup";
		else
			radioGroupCountText = "#RadioGroup = " + radioGroupCount;

		if(listPrefCount ==0 ) 
			listPrefCountText = "no ListPref";
		else
			listPrefCountText = "#ListPref = " + listPrefCount;

		if(seekBarCount ==0 ) 
			seekBarCountText = "no SeekBar";
		else
			seekBarCountText = "#SeekBar = " + seekBarCount;

		if(numberPickerCount ==0 ) 
			numberPickerCountText = "no NumberPicker";
		else
			numberPickerCountText = "#NumberPicker = " + numberPickerCount;

		String template =
				"module " + moduleName + END_LINE + 
				END_LINE + "open interfaceModel"+ END_LINE +
				END_LINE + formalizedModel.toString() + END_LINE + END_LINE + 
				"pred show{" + END_LINE +
				TAB + checkBoxCountText + END_LINE +
				TAB + textEditCountText + END_LINE +
				TAB + spinnerCountText + END_LINE +
				TAB + datePickerCountText + END_LINE +
				TAB + radioGroupCountText + END_LINE +
				TAB + textEditPrefCountText + END_LINE +
				TAB + checkBoxPrefCountText + END_LINE +				
				TAB + listPrefCountText + END_LINE +				
				TAB + numberPickerCountText + END_LINE +
				TAB + seekBarCountText + END_LINE +
				"}" + END_LINE +
				"run show";
		
		if (fileName.contains(".xml"))
			fileName = fileName.replace(".xml", "");

		writeToFile(template, fileName);
	}

	// generate alloy for all combinations , TESTED
	private void generateAllCombinations(String fileName) {
		String seekBarCountText;
		String numberPickerCountText;
		String checkBoxCountText;
		String textEditCountText;
		String spinnerCountText;
		String datePickerCountText;
		String radioGroupCountText;
		String checkBoxPrefCountText;
		String textEditPrefCountText;
		String listPrefCountText;
		String moduleName = fileName.replace("$$$", "_");
		moduleName = moduleName.replace(".xml", "_");
		moduleName = moduleName.replace("#", "_");
		int seekBarCount = 0, numberPickerCount = 0, checkBoxCount = 0, textEditCount = 0 , checkBoxPrefCount = 0, listPrefCount = 0, textEditPrefCount = 0 , spinnerCount = 0 , datePickerCount = 0 , radioGroupCount = 0;

		StringBuffer formalizedModel = new StringBuffer();

		for (Map.Entry<String, String> entry : formalizedModelMap.entrySet()) {
			formalizedModel.append(entry.getValue());

			// if widget is not on screen continue
			if(!interfaceModel.contains(entry.getKey()))
				continue;

			String type = interfaceModel.getIdToType(entry.getKey());

			if(type.equals(EditText.TYPE)) 
				textEditCount ++;
			if(type.equals(RadioGroup.TYPE)) 
				radioGroupCount ++;
			if(type.equals(DatePicker.TYPE)) 
				datePickerCount ++;
			if(type.equals(Spinner.TYPE)) 
				spinnerCount ++;
			if(type.equals(CheckBox.TYPE)) 
				checkBoxCount ++;
			if(type.equals(EditTextPreference.TYPE)) 
				textEditPrefCount ++;
			if(type.equals(CheckBoxPreference.TYPE)) 
				checkBoxPrefCount ++;
			if(type.equals(ListPreference.TYPE)) 
				listPrefCount ++;
			if(type.equals(SeekBar.TYPE)) 
				seekBarCount ++;
			if(type.equals(NumberPicker.TYPE)) 
				numberPickerCount ++;
		}

		if(checkBoxCount ==0 ) 
			checkBoxCountText = "no CheckBox";
		else
			checkBoxCountText = "#CheckBox = " + checkBoxCount;

		if(textEditCount ==0 ) 
			textEditCountText = "no TextEdit";
		else
			textEditCountText = "#TextEdit = " + textEditCount;

		if(checkBoxPrefCount ==0 ) 
			checkBoxPrefCountText = "no CheckBoxPref";
		else
			checkBoxPrefCountText = "#CheckBoxPref = " + checkBoxPrefCount;

		if(textEditPrefCount ==0 ) 
			textEditPrefCountText = "no TextEditPref";
		else
			textEditPrefCountText = "#TextEditPref = " + textEditPrefCount;

		if(spinnerCount ==0 ) 
			spinnerCountText = "no Spinner";
		else
			spinnerCountText = "#Spinner = " + spinnerCount;

		if(datePickerCount ==0 ) 
			datePickerCountText = "no DatePicker";
		else
			datePickerCountText = "#DatePicker = " + datePickerCount;

		if(radioGroupCount ==0 ) 
			radioGroupCountText = "no RadioGroup";
		else
			radioGroupCountText = "#RadioGroup = " + radioGroupCount;


		if(listPrefCount ==0 ) 
			listPrefCountText = "no ListPref";
		else
			listPrefCountText = "#ListPref = " + listPrefCount;

		if(seekBarCount ==0 ) 
			seekBarCountText = "no SeekBar";
		else
			seekBarCountText = "#SeekBar = " + seekBarCount;

		if(numberPickerCount ==0 ) 
			numberPickerCountText = "no NumberPicker";
		else
			numberPickerCountText = "#NumberPicker = " + numberPickerCount;
		
		String template =
				"module " + moduleName + END_LINE + 
				END_LINE + "open interfaceModel"+ END_LINE +
				END_LINE + formalizedModel.toString() + END_LINE + END_LINE + 
				"pred show{" + END_LINE +
				TAB + checkBoxCountText + END_LINE +
				TAB + textEditCountText + END_LINE +
				TAB + spinnerCountText + END_LINE +
				TAB + datePickerCountText + END_LINE +
				TAB + radioGroupCountText + END_LINE +
				TAB + textEditPrefCountText + END_LINE +
				TAB + checkBoxPrefCountText + END_LINE +
				TAB + listPrefCountText + END_LINE +
				TAB + numberPickerCountText + END_LINE +
				TAB + seekBarCountText + END_LINE +
				"}" + END_LINE +
				"run show";
		
		if (fileName.contains(".xml"))
			fileName = fileName.replace(".xml", "");
		writeToFile(template, fileName);
	}

	private void writeToFile(String output, String fileName) {
		try {
			File file = new File(outputPath);

			if (!file.exists()) {
				if (file.mkdirs()) {
					System.out.println("Directory is created!");
				} else {
					System.out.println("Failed to create directory!");
				}
			}

			FileWriter fstream = new FileWriter(outputPath+fileName+".als");
			BufferedWriter buffOut = new BufferedWriter(fstream);

			System.out.println("Generating...");
			System.out.println();


			buffOut.append(output);					
			buffOut.close();

		} catch (Exception e) {
			e.printStackTrace();
		}	
	}

	public LayoutToAlloy(String targetApp, String outputPath, String stringsXml, DependencyExtractor dependencyExtractor, boolean prunning) {
		this.targetApp = targetApp;
		this.outputPath = outputPath;
		this.stringsXml = stringsXml;
		this.prunning = prunning;
		this.dependencyExtractor = dependencyExtractor;
		
		File folder = new File(targetApp+"/res/layout");
		File[] listOfFiles = folder.listFiles();
				
		if(listOfFiles != null)
			for (int i = 0; i < listOfFiles.length; i++) {
				if (listOfFiles[i].isFile()) {
					parseLayout(listOfFiles[i].getPath(), listOfFiles[i].getName().split("[.]")[0], true);
				} 
			}

	}

	public LayoutToAlloy(String targetApp, String targetLayoutName, String targetLayoutPath, String outputPath, String stringsXml, DependencyExtractor dependencyExtractor, boolean prunning) {
		this.targetApp = targetApp;
		this.outputPath = outputPath;
		this.stringsXml = stringsXml;
		this.dependencyExtractor = dependencyExtractor;
		this.prunning = prunning;

		parseLayout(targetApp + targetLayoutPath, targetLayoutName, false);	
	}

	//TODO: improve this
	// return a list of all possible actions that have a defined handler on the activity
	private List<String> listActivityActions(Document document){
		List<String> actions = new ArrayList<>();
		NodeList buttons = document.getElementsByTagName("Button");

		if(buttons != null && buttons.getLength() > 0) {
			for(int j = 0 ; j < buttons.getLength();j++) {

				Element el = (Element)buttons.item(j);
				if (el.getAttribute("android:id") == null)
					continue;

				// TODO: add a way to track widgets that don't have an ID
				String id = el.getAttribute("android:id");
				if(id == null || id.isEmpty() || id.trim().equals("")) {
					id = el.getAttribute("android:onClick");
				} else
					id = id.split("/")[1].trim();					


				actions.add(id);
			}
		} 
		return actions;
	}

	public LayoutMetaData getInterfaceModel() {
		return interfaceModel;
	}	
}
