package edu.gmu.trimdroid.testgen;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.gmu.trimdroid.metadata.InterfaceModel;
import edu.gmu.trimdroid.soot.utils.Util;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import edu.gmu.trimdroid.AbstractXmlParser;
import edu.gmu.trimdroid.alloy.Evaluator;
import edu.gmu.trimdroid.alloy.LayoutToAlloy;
import edu.gmu.trimdroid.metadata.ActivityMetaData;
import edu.gmu.trimdroid.metadata.RadioGroupMetaData;
import edu.gmu.trimdroid.soot.depextract.DependencyExtractor;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;

public class TestCaseGenerator extends AbstractXmlParser {

	private static final String END_LINE = "\n";
	private static final String TAB = "\t";	
	
	private String srcPath;
	private int count=0;
	private List<List<String>> paths;
	private Map<String,List<String>> activityToInputs = new HashMap<String, List<String>>();
	private Map<String,ActivityMetaData> activities;
	private InterfaceModel interfaceModel;
	private String rFilePackage;
	private boolean sourceAvailable;
	private String appName;
	private boolean eliminateDuplicates = false;
	private Set<String> inputsHashSet = new HashSet<String>();
	
	public TestCaseGenerator(String srcPath, String appName, String imSolutions, String bmSolutions, DependencyExtractor dependencyExtractor, boolean sourceAvailable, boolean prunning) throws ParserConfigurationException, SAXException, IOException {
		this.appName = appName;
		this.srcPath = srcPath;
		this.sourceAvailable = sourceAvailable;
		File folder;
		if (sourceAvailable) {
			interfaceModel = new InterfaceModel(srcPath);
			folder = new File("out/"+appName+imSolutions);
		}
		else {
			interfaceModel = new InterfaceModel(srcPath);
			folder = new File(srcPath+imSolutions);			
		}
			
		this.rFilePackage = extractRFilePackage(srcPath);
		
		activities = interfaceModel.getActivities();

		this.paths = generateBMPaths(bmSolutions);
		
		File[] listOfFiles = folder.listFiles();

		if(listOfFiles != null)
			for (int i = 0; i < listOfFiles.length; i++) {
				if (listOfFiles[i].isFile()) {
					if(listOfFiles[i].toString().contains(".xml")) {
						generateRobootiumInputs(listOfFiles[i].getPath(), listOfFiles[i].getName(), null);
					}
				} 
			}
	
		// create a test case for each path 
		for(List<String> p:this.paths) {
			String startActivity = "";

			if(p.isEmpty())
				continue;

			if(p.get(0).contains("____"))
				startActivity = p.get(0).split("____")[0];
			else 
				startActivity = p.get(0).split("___")[0];

			String packageName = activities.get(startActivity).getPackageName();
			String activityName = startActivity;
			String handlerMethod;		
			String perviousHandler = "";		
			ActivityMetaData activityMetaData;
			List<List<String>> pathInputs = new ArrayList<List<String>>();
			int maxSize = 0;
			int index = 0;
			String node;

			List<String> path = p;
			for(int j=0; j<path.size(); j++) {
				node = path.get(j);
				if(node.contains("____"))
					handlerMethod = node.split("___")[2];
				else 
					handlerMethod = node.split("___")[1];

				String previousActivity = "";
				String nextHanlder = "";

				if(j>0) {
					previousActivity = path.get(j-1);  
					
					if(previousActivity.contains("____")) {
						perviousHandler = previousActivity.split("___")[2];
						previousActivity = previousActivity.split("____")[0];
					}
					else {
						perviousHandler = previousActivity.split("___")[1];					
						previousActivity = previousActivity.split("___")[0];
					}
				}
				if(j+1<path.size()) {
					String nextActivity = path.get(j+1);  
					if(nextActivity.contains("____"))
						nextHanlder = nextActivity.split("___")[2];
					else 
						nextHanlder = nextActivity.split("___")[1];		
				}

				String newActivity = "";
				System.out.println(path.get(j));
				
				if (j == 1 ) {
					if (handlerMethod.equals("onItemLongClick") && perviousHandler.equals("onCreate"))
						break;
					if (handlerMethod.equals("onCreate") && path.get(j).split("___")[0].equals("BeerInfo")) {
						List<String> constructedPath;
						String checkOption = END_LINE + TAB + TAB + "solo.clickInList(1);" + END_LINE;
						if (pathInputs.isEmpty()) {
							constructedPath = new ArrayList<>();
							constructedPath.add(checkOption);
							pathInputs.add(constructedPath);					
						} else {
							constructedPath = pathInputs.get(pathInputs.size()-1);
							for (int m =0; m < constructedPath.size(); m++) {
								constructedPath.set(m, constructedPath.get(m) + checkOption);
							}
						}
					}
				}
				
				if (handlerMethod.equals("onCheckedChanged") && perviousHandler.equals("initConfigDialog")) {
					String checkOption = "";
					String optionID = "";
					int i = 2;
					if (node.contains("____"))
						i = 3;

					if (node.split("___").length > 1)
						optionID = node.split("___")[i];

					if (optionID != null && !optionID.equals("")) {
						checkOption = END_LINE + TAB + TAB + "CheckBox cb_"+optionID+" = (CheckBox) solo.getView(R.id."+optionID+");" + END_LINE;
						checkOption += TAB + TAB + "solo.clickOnView(cb_"+optionID+");" + END_LINE;
					}
					List<String> constructedPath = pathInputs.get(pathInputs.size()-1);
					for (int m =0; m < constructedPath.size(); m++) {
						constructedPath.set(m, constructedPath.get(m) + checkOption);
					}
				}
				
				if (handlerMethod.equals("onListItemClick")) {				
					String checkOption = "";
					String optionID = "";
					int i = 2;
					if (node.contains("____"))
						i = 3;

					if (node.split("___").length > 1)
						optionID = node.split("___")[i];

					if (optionID != null && !optionID.equals("") && !optionID.equals("IPC")) {
						checkOption = END_LINE + TAB + TAB + "solo.clickInList("+optionID+");" + END_LINE;
					} else {
						if (path.size() > j+1 && path.get(j+1) != null && !path.get(j+1).equals("") && !path.get(j+1).split("___")[1].equals("onListItemClick")) {
							i = 2;
							if (path.get(j+1).contains("____"))
								i = 3;

							if (path.get(j+1).split("___").length > 1)
								optionID = path.get(j+1).split("___")[i];
								int itemIndex = Integer.parseInt(optionID) + 1;
							checkOption = END_LINE + TAB + TAB + "solo.clickInList("+ itemIndex +");" + END_LINE;							
						}
					}
					List<String> constructedPath = pathInputs.get(pathInputs.size()-1);
					for (int m =0; m < constructedPath.size(); m++) {
						constructedPath.set(m, constructedPath.get(m) + checkOption);
					}
				}
				
				if( handlerMethod.equals("onCreate") && !(perviousHandler+previousActivity).equals(handlerMethod+activityName) && ! nextHanlder.equals("onClick") ) {
					if(node.contains("___Loop___"))
						newActivity = previousActivity;
					else if(p.get(j).contains("____"))
						newActivity = node.split("____")[0];
					else 
						newActivity = node.split("___")[0];

					String actionRobotium = "";

					//TODO: FINISH This if there's no a child node
					if(j+1 == p.size() || (!p.get(j+1).contains("onClick"))) {

						activityMetaData = activities.get(newActivity);
						
						if(perviousHandler.equals("onMenuItemSelected") || perviousHandler.equals("onOptionsItemSelected")) {
							handlerMethod = perviousHandler;

							if(activities.get(newActivity) != null)
								activityName = newActivity;

							activityMetaData = activities.get(activityName);

							String widgetID = "";

							int i = 2;
							if (node.contains("____"))
								i = 3;

							if (node.split("___").length > 1)
								widgetID = node.split("___")[i];


							// TODO: handle special cases of widget ID
							if(widgetID == null || widgetID.equals(""))
								break;

							actionRobotium = generateAction(handlerMethod, widgetID);

						}
						
						// TODO: fix this to ignore services
						if(newActivity.contains("Service") || activityMetaData == null)
							continue;
						String layoutName = activityMetaData.getLayoutName();
						// if the activity does not have a layout file
						if ( layoutName == null)
							continue;
						if (layoutName.endsWith(".xml"))
							layoutName = layoutName.replace(".xml", "").trim();


						List<String> robotiumInputs = new ArrayList<String>();
						List<String> pathRobotiums = new ArrayList<String>();

						if(activityToInputs.get(layoutName) != null && !activityToInputs.get(layoutName).isEmpty()) {
							robotiumInputs = activityToInputs.get(layoutName);
						}	
						else {
							String imPath;
							if (sourceAvailable)
								imPath = "out/"+ appName + "/im/";
							else 
								imPath = srcPath + "/im/";
							
							String alloyFileName = imPath + layoutName + ".als";
							File alloyFile = new File( alloyFileName);

							if(!alloyFile.exists()) {
								new LayoutToAlloy(srcPath, layoutName, activityMetaData.getLayoutPath(), "out/"+appName+"/im/", srcPath + "/res/values/strings.xml", dependencyExtractor, prunning);
							}
							
							//TODO: to improve performance only run the evaluator for new alloy files
							new Evaluator(alloyFile);
							
							if (sourceAvailable)
								folder = new File("out/" + appName + imSolutions+"/" + layoutName.replace(".xml", ""));
							else
								folder = new File(srcPath + imSolutions + "/" + layoutName.replace(".xml", ""));
								
							listOfFiles = folder.listFiles();

							if(listOfFiles != null)
								for (int i = 0; i < listOfFiles.length; i++) {
									if (listOfFiles[i].isFile()) {
										if(listOfFiles[i].toString().contains(".xml")) {
											generateRobootiumInputs(listOfFiles[i].getPath(), listOfFiles[i].toString() , layoutName);
										}
									} 
								}							
						}

						if(activityToInputs.get(layoutName) != null && !activityToInputs.get(layoutName).isEmpty()) {
							robotiumInputs = activityToInputs.get(layoutName);
						}	

						// if action string is empty or null go to next path
						if (actionRobotium != null && !actionRobotium.trim().equals("")) {
														
							for(String r:robotiumInputs) {
								pathRobotiums.add(actionRobotium + END_LINE + r);
							} 

							if (robotiumInputs.isEmpty())
								pathRobotiums.add(actionRobotium);

							pathInputs.add(pathRobotiums);

							if(pathRobotiums.size() > maxSize) {
								maxSize = robotiumInputs.size();
							}		

						} else {

							if(!robotiumInputs.isEmpty()) {
								pathInputs.add(robotiumInputs);

								if(robotiumInputs.size() > maxSize) {
									maxSize = robotiumInputs.size();
								}					
							}
						}	
					}

					continue;
				}

				if(!handlerMethod.equals("onClick") && (perviousHandler.equals("onMenuItemSelected") || perviousHandler.equals("onOptionsItemSelected")))
					handlerMethod = perviousHandler;

				if(activities.get(newActivity) != null)
					activityName = newActivity;

				activityMetaData = activities.get(activityName);

				String widgetID = "";

				int i = 2;
				if (node.contains("____"))
					i = 3;

				if (node.split("___").length > 1) {
					try {
						widgetID = node.split("___")[i];
					}
					catch(Exception e) {
						widgetID = node.split("___")[i-1];
					}
				}

				// TODO: handle special cases of widget ID
				if(widgetID == null || widgetID.equals(""))
					break;


				// if last node is onClick
				if (handlerMethod.equals("onCreate")) {
					if(p.get(j).contains("____"))
						newActivity = node.split("____")[0];
					else 
						newActivity = node.split("___")[0];
					if(activities.get(newActivity) == null)
						newActivity = activityName;
					else 
						activityName = newActivity;
				}
				
				if(widgetID.equals("SYMBOLIC") || widgetID.equals("IPC") || widgetID.equals("SIMPLE"))
					continue;


				String optionText = "";
				if(widgetID.contains("LLLLLListOption"))
					optionText = TAB + TAB + "solo.clickInList(0);" + END_LINE + TAB + TAB + "solo.sleep(4000);" + END_LINE;

				if(widgetID.contains("PPPPPPositiveBtn"))
					optionText = TAB + TAB + "solo.enterText(0,\"1\");" + END_LINE + TAB + TAB + "solo.clickOnButton(0);" + END_LINE + TAB + TAB +"solo.sleep(4000);" + END_LINE;
				if(widgetID.contains("NNNNNNegativeBtn"))
					optionText = TAB + TAB + "solo.enterText(0,\"1\");" + END_LINE + TAB + TAB + "solo.clickOnButton(1);" + END_LINE + TAB + TAB +"solo.sleep(4000);" + END_LINE;
				
				List<String> robotiumInputs = new ArrayList<String>();
				List<String> pathRobotiums = new ArrayList<String>();
				String layoutName = activityMetaData.getLayoutName().trim();
				if (layoutName.endsWith(".xml"))
					layoutName = layoutName.replace(".xml", "").trim();
				
				if(optionText.equals("") && activityToInputs.get(layoutName+"$$$"+widgetID) != null && !activityToInputs.get(layoutName+"$$$"+widgetID).isEmpty()) {
					robotiumInputs = activityToInputs.get(layoutName+"$$$"+widgetID);
				}	
				else {
					// TODO: call LayoutToAlloy for the layout and action
					robotiumInputs = new ArrayList<String>();
				}
				String actionRobotium = "";
				if (optionText.equals("") )
					actionRobotium = generateAction(handlerMethod, widgetID);

				// if action string is empty or null go to next path
				if ((actionRobotium == null || actionRobotium.trim().equals("")) && optionText.equals(""))
					continue;

				for(String r:robotiumInputs) {
					pathRobotiums.add(r + actionRobotium);
				} 


				if (robotiumInputs.isEmpty())
					pathRobotiums.add(actionRobotium + optionText);

				pathInputs.add(pathRobotiums);

				if(pathRobotiums.size() > maxSize) {
					maxSize = pathRobotiums.size();
				}				

			}

			// concatenate inputs in the path
			for (int k = 0; k < maxSize; k++) {
				StringBuffer testCaseData = new StringBuffer();

				for (int j = 0; j < pathInputs.size(); j++) {

					index = k % pathInputs.get(j).size();
					testCaseData.append(pathInputs.get(j).get(index));
					testCaseData.append(END_LINE);
				}	
				
				generateRobotiumTestCase(startActivity,packageName, startActivity+"Test_"+count,testCaseData.toString());
			}
		}

		System.out.println("Finished Generating Inputs...");
	}

	private String generateAction(String handlerMethod, String widgetID) {
		String actionString;
		if(widgetID == null)
			return null;

		if(handlerMethod.equals("onClick")) {
			if (Util.isNumeric(widgetID)) {
				actionString = TAB + TAB + "solo.clickOnText(\"OK\");" + END_LINE;
			} else {
				if (widgetID.equals("i4"))
					return null;
				actionString = TAB + TAB + "View v_"+widgetID+" = solo.getView(R.id."+widgetID+");" + END_LINE;
				actionString += TAB + TAB + "solo.clickOnView(v_"+ widgetID + ");";
			}
			//			actionString = TAB + TAB + "solo.clickOnButton(R.id."+widgetID+");";
			return actionString;
		}
		if(handlerMethod.equals("onMenuItemSelected") || handlerMethod.equals("onOptionsItemSelected")) {
			if (widgetID.equals("optionsmenu_statistics"))
				return null;
			try{
				int intWidgetId = Integer.parseInt(widgetID);
				if(intWidgetId < 20)
					actionString = TAB + TAB + "solo.pressMenuItem("+intWidgetId+");";
				else {
					actionString = TAB + TAB + "solo.clickOnActionBarItem(R.id."+widgetID+");";
				}
					
			}catch(NumberFormatException e){
				if (!widgetID.equals("SYMBOLIC"))
					actionString = TAB + TAB + "solo.clickOnActionBarItem(R.id."+widgetID+");";
				else
					actionString = null;
			}
			return actionString;
		}

		return null;	
	}

	private void generateRobootiumInputs(String inputXMLPath, String xmlFileName, String layoutName) {

		Document document = prepareXML(inputXMLPath);
		StringBuffer robotiumInputData = new StringBuffer();
		String addOn = "";
		if (document!=null) {
			NodeList fields = document.getElementsByTagName("field");

			if(fields != null && fields.getLength() > 0) {
				for(int i = 0 ; i < fields.getLength();i++) {
					NodeList tuples = ((Element)fields.item(i)).getElementsByTagName("tuple");

					if(tuples != null && tuples.getLength() > 0) {
						for(int j = 0 ; j < tuples.getLength();j++) {
							NodeList atoms = ((Element)tuples.item(j)).getElementsByTagName("atom");
							if(atoms != null && atoms.getLength() > 0) {
								// widget ID
								String widgetID = ((Element)atoms.item(0)).getAttribute("label");

								String widgetType;
								String[] strArr = widgetID.split("_"); 
								if ( strArr.length >2) {
									widgetType = (strArr[strArr.length-1].trim()).split("[$]")[0];
									if(widgetID.contains("$"))
										widgetID = widgetID.split("[$]")[0];

									widgetID = widgetID.substring(0,widgetID.length() - widgetType.length() - 1);
								}
								else {
									widgetType = (strArr[1].trim()).split("[$]")[0];
									widgetID = strArr[0].trim();
								}
								// value
								String value = ((Element)atoms.item(1)).getAttribute("label");
								if(value.contains("/"))
									value = value.split("/")[1].trim().split("[$]")[0];

								String activityName;
								if(layoutName == null) {
									String extractedLayoutName = "";
									if(xmlFileName.contains("$$$")) {
										extractedLayoutName = xmlFileName.split("\\$\\$\\$")[0];
										activityName = interfaceModel.getActivityFromLayout(extractedLayoutName);
									}
									else {
										extractedLayoutName = xmlFileName.split("___")[0];										
										activityName = interfaceModel.getActivityFromLayout(extractedLayoutName);
									}
								}
								else {
									activityName = interfaceModel.getActivityFromLayout(layoutName);
								}
								String temp = transformToRobotium(widgetType, widgetID, value, activityName);
								
								if (temp.contains("Show rating reminders"))
									addOn = temp;
								else
									if (widgetType.equals("CheckBoxPref") && activityName.equals("Config")) {
										robotiumInputData.insert(0, temp);
									}
									else
										robotiumInputData.append(temp);

							}
						}
					}	
				}
			}
			// key is in format of layout_action
			String key;

		
			if(xmlFileName.contains("___"))
				xmlFileName = xmlFileName.split("___")[0];
			key = xmlFileName;

			if(xmlFileName.contains("$$$"))
				xmlFileName = xmlFileName.split("\\$\\$\\$")[0] + xmlFileName.split("\\$\\$\\$")[1];

			if(layoutName != null)
				key = layoutName;
			List<String> combinations = activityToInputs.get(key);
			if(combinations == null) 
				combinations = new ArrayList<String>();

			
			combinations.add(addOn + robotiumInputData.toString());
			activityToInputs.put(key,combinations);
		}
	}

	private List<List<String>> generateBMPaths(String   bmSolutions) {
		File folder;
		if (this.sourceAvailable)
			folder = new File("out/"+appName+bmSolutions);
		else
			folder = new File(srcPath+bmSolutions);
			
		File[] listOfFiles = folder.listFiles();

		List<List<String>> paths = new ArrayList<List<String>>();

		for (int i = 0; i < listOfFiles.length; i++) {
			if (listOfFiles[i].isFile()) {
				if(listOfFiles[i].toString().contains(".xml")) {
					List<String> path = generatePathFromSolution(listOfFiles[i].getPath());
					paths.add(path);
				}
			} 
		}
		return paths;
	}

	private List<String> generatePathFromSolution(String solvedBMLPath) {
		Document document = prepareXML(solvedBMLPath);
		List<String> pathActions = new ArrayList<String>();

		if (document!=null) {
			NodeList fields = document.getElementsByTagName("field");

			if(fields != null && fields.getLength() > 0) {
				for(int i = 0 ; i < fields.getLength();i++) {
					String label = ((Element)fields.item(i)).getAttribute("label");

					if(label != null && label.equals("edges")) {
						NodeList tuples = ((Element)fields.item(i)).getElementsByTagName("tuple");

						for(int j = 0 ; j < tuples.getLength();j++) {
							NodeList atoms = ((Element)tuples.item(j)).getElementsByTagName("atom");
							if(atoms != null && atoms.getLength() > 0) {

								String action = ((Element)atoms.item(1)).getAttribute("label");
								action = (action.trim()).split("[$]")[0];
								if(action.equals("Root"))
									continue;
								String handlerMethod = (action.trim()).split("___")[1];
								String activity = (action.trim()).split("__")[0];

								//								String onScreenRobotiumAction = processHandler(activity,handlerMethod);
								//								
								//								if(onScreenRobotiumAction != "")
								//									pathActions.add(activity+"_"+onScreenRobotiumAction);
								pathActions.add(action);
								//								System.out.println(action);
								if(j == tuples.getLength()-1 ) {
									action = ((Element)atoms.item(2)).getAttribute("label");
									action = (action.trim()).split("[$]")[0];

									handlerMethod = (action.trim()).split("___")[1];
									activity = (action.trim()).split("__")[0];
									//									onScreenRobotiumAction = processHandler(activity,handlerMethod);
									//									
									//									if(onScreenRobotiumAction != "")
									//										pathActions.add(activity+"_"+onScreenRobotiumAction);
									pathActions.add(action);
								}

							}
						}
					}	
				}
			}
		}
		return pathActions;
	}

	private String transformToRobotium(String type, String id, String value, String activityName) {
		String inputString;
		ActivityMetaData activityMetaData = activities.get(activityName);

		if(type.equals("TextEdit")) {
			inputString = TAB + TAB + "EditText et_"+id+" = (EditText) solo.getView(R.id."+id+");" + END_LINE;
			value = "1";
			if(value.equals("numberDecimal")) {
				value = "83.5";
			} else if(value.equals("integer")) {
				value = "7";
			}
			inputString += TAB + TAB + "solo.clearEditText(et_"+id+");" + END_LINE;			
			inputString += TAB + TAB + "solo.enterText(et_"+id+", \""+ value +"\");" + END_LINE;			
			return inputString;
		}		
		if(type.equals("Spinner")) {
			
			inputString = 				TAB + TAB + "try { "
					+ "this.runTestOnUiThread(new Runnable()  {" + END_LINE
					+ "@Override" + END_LINE
					+ "public void run() {"+ END_LINE ;
			
			inputString += TAB + TAB + "Spinner sp_"+id+" = (Spinner) solo.getView(R.id."+id+");" + END_LINE;
			int choice = Integer.parseInt(value)+8;
//			LayoutMetaData layoutMetaData = new LayoutMetaData(srcPath+activityMetaData.getLayoutPath(), srcPath + "/res/values/strings.xml");			
			// always increase choice by one
//			List<Spinner> spinners = layoutMetaData.getSpinners();
//			for (Spinner s:spinners) {
//				if (s.getId().equals(id)) {
//					choice = (choice + 1) % s.getSpinnerItemsCount();
//					break;
//				}
//			}
			inputString += TAB + TAB + "sp_"+id+".setSelection("+ choice +", true);" + END_LINE;
			
			inputString += 				TAB + TAB + "}" +
					TAB + TAB + " }); "+ END_LINE +
					TAB + TAB + "} catch (Throwable e) {"+ END_LINE +
					TAB + TAB + "	e.printStackTrace();"+ END_LINE +
					TAB + TAB + "}"+ END_LINE ; 
			return inputString;
		}
		if(type.equals("CheckBox")) {
			if(value.equals("checked")) {

				inputString = TAB + TAB + "CheckBox cb_"+id+" = (CheckBox) solo.getView(R.id."+id+");" + END_LINE;
				inputString += TAB + TAB + "if(cb_"+id+".isChecked()) {" + END_LINE;
				inputString += TAB + TAB + TAB + "solo.clickOnView(cb_"+id+");" + END_LINE;
				inputString += TAB + TAB + "}" + END_LINE;
				return inputString;
			}
			else {
				inputString = TAB + TAB + "CheckBox cb_"+id+" = (CheckBox) solo.getView(R.id."+id+");" + END_LINE;
				inputString += TAB + TAB + "if(!cb_"+id+".isChecked()) {" + END_LINE;
				inputString += TAB + TAB + TAB + "solo.clickOnView(cb_"+id+");" + END_LINE;
				inputString += TAB + TAB + "}" + END_LINE;
				return inputString;				
			}
		}
		if(type.equals("RadioGroup")) {		
			String radioButtonId = activityMetaData.getRadioGroupById(id).getRadioButtonsIds().get(Integer.parseInt(value)+8);

			inputString = TAB + TAB + "RadioButton rb_"+radioButtonId+" = (RadioButton) solo.getView(R.id."+radioButtonId+");" + END_LINE;
			inputString += TAB + TAB + "solo.clickOnView(rb_"+radioButtonId+");" + END_LINE;
			return inputString;
		}
		if(type.equals("DatePicker")) {

		}

		if(type.equals("TextEditPref")) {
			value = "1";

			if(value.equals("numberDecimal") || value.equals("integer")) {
				value = "5";
			}
			inputString = TAB + TAB + "solo.clickOnText(\""+extractTitle(id)+"\");" + END_LINE;			
			inputString += TAB + TAB + "solo.clearEditText(0);" + END_LINE;			
			inputString += TAB + TAB + "solo.enterText(0, \""+ value +"\");" + END_LINE;			
			inputString += TAB + TAB + "solo.clickOnText(\"OK\");" + END_LINE;			
			return inputString;
		}
		if(type.equals("SeekBar")) {
			String temp = extractTitle(id).replaceAll("\"", "");
			inputString = TAB + TAB + "solo.clickOnText(\""+temp+"\");" + END_LINE;
			inputString += TAB + TAB + "solo.clickOnText(\"OK\");" + END_LINE;
			return inputString;
		}
		if(type.equals("NumberPicker")) {
			String temp = extractTitle(id).replaceAll("\"", "");
			inputString = TAB + TAB + "solo.clickOnText(\""+temp+"\");" + END_LINE;
			inputString += TAB + TAB + "solo.enterText(0,\"3\");" + END_LINE;
			inputString += TAB + TAB + "solo.clickOnText(\"OK\");" + END_LINE;
			return inputString;
		}
		if(type.equals("ListPref")) {
			int choice = Integer.parseInt(value)+8;
			String temp = extractTitle(id).replaceAll("\"", "");
			inputString = TAB + TAB + "solo.clickOnText(\""+temp+"\");" + END_LINE;			
			inputString += TAB + TAB + "solo.clickInList("+choice+");" + END_LINE;			
			return inputString;
		}
		if(type.equals("CheckBoxPref")) {
			inputString = "";
			if(value.equals("checked")) {
				String temp = extractTitle(id).replaceAll("\"", "");
				inputString = TAB + TAB + "solo.clickOnText(\""+temp+"\");" + END_LINE;			
			}
			return inputString;
		}
		return null;
	}

	private void generateRobotiumTestCase(String activityName,
			String packageName, String testClassName, String inputs) {

		String prep = "";
		if (activityName.equals("DivideAndConquerActivity")) 
			prep = "solo.clickOnText(\"New game\");";
		
		String template =

				"package _PACKAGE_NAME_;" + END_LINE + 
				"import android.test.ActivityInstrumentationTestCase2;"+ END_LINE +
				"import android.test.suitebuilder.annotation.Smoke;"+ END_LINE +
				"import com.robotium.solo.Solo;"+ END_LINE +
//				"import com.jayway.android.robotium.solo.Solo;"+ END_LINE +
				"import android.widget.RadioButton;"+ END_LINE +
				"import android.widget.CheckBox;"+ END_LINE +
				"import android.widget.EditText;"+ END_LINE +
				"import android.widget.Spinner;"+ END_LINE +
				"import android.view.View;"+ END_LINE +
				"import _ACTIVITY_PACKAGE_NAME_._ACTIVITY_CLASS_NAME_;" +  END_LINE +
				"import "+ this.rFilePackage +".R;" +  END_LINE +

				"public class _CLASS_NAME_ extends ActivityInstrumentationTestCase2<_ACTIVITY_CLASS_NAME_>"+ END_LINE +
				"{"+ END_LINE +
				TAB + "private Solo solo;"+ END_LINE + END_LINE +

				TAB + "public _CLASS_NAME_()   {"+ END_LINE +
				TAB + TAB + "super(\"_ACTIVITY_PACKAGE_NAME_\",_ACTIVITY_CLASS_NAME_.class);"+ END_LINE +
				TAB + "};"+ END_LINE + END_LINE +

				TAB + "public void setUp() throws Exception   {" + END_LINE +
				TAB + TAB + "solo = new Solo(getInstrumentation(), getActivity());"+ END_LINE +
				TAB + TAB + prep + END_LINE + 
				TAB + "};"+ END_LINE + END_LINE +

				TAB + "@Smoke"+ END_LINE +
				TAB + "public void testMethod() throws Exception   {"+ END_LINE +
				"_POPULATE_INPUT_BOXES_"+ END_LINE +
				TAB + TAB + "solo.sleep(2000);"+ END_LINE +						
				TAB + "};"+ END_LINE + END_LINE +

				TAB + "@Override"+ END_LINE +
				TAB + "public void tearDown() throws Exception {"+ END_LINE +
				TAB + TAB + "solo.finishOpenedActivities();"+ END_LINE +
				TAB + "};"+ END_LINE + END_LINE +

				"};";

		
		String testCase = template;


		//TODO: FIX THIS
		testCase = testCase.replaceAll("_ACTIVITY_PACKAGE_NAME_", packageName);
		testCase = testCase.replaceAll("_ACTIVITY_CLASS_NAME_", activityName);
		testCase = testCase.replaceAll("_PACKAGE_NAME_", packageName+".test");
		testCase = testCase.replaceAll("_CLASS_NAME_", testClassName);	

		// move the cursor to the first item in the list
		String listOption = "";
		if (activities.get(activityName).hasListView() && srcPath.endsWith("httpmon")) {
			listOption = END_LINE + TAB + TAB + "solo.scrollDownList(0);" + END_LINE;
		}
		
		String inputString = listOption + inputs;	
		
		// eliminate duplicate test cases
		eliminateDuplicates = enableFiltering(packageName);
		if (eliminateDuplicates && inputsHashSet.contains(inputString)) {
			return;
		} else {
			inputsHashSet.add(inputString);
		}
		
		// check input string for duplicate field definition
		inputString = removeDuplicateFields(inputString);
		
		testCase = testCase.replaceAll("_POPULATE_INPUT_BOXES_", inputString);

		//		String buttonCickString = "solo.clickOnButton(\""+action+"\");";
		//		testCase = testCase.replaceAll("_CLICK_BUTTON_", buttonCickString);
		count++;

		writeToFile(testCase,testClassName +".java");		
	}

	private String removeDuplicateFields(String inputString) {
		String[] lines= inputString.split(END_LINE);
		Set<String> temp = new HashSet<>();
		StringBuilder result = new StringBuilder();
		for (int i = 0; i < lines.length; i++) {
			String line = lines[i].trim();
			if (temp.contains(line)) {
				if (line.startsWith("EditText ") || line.startsWith("View ")) {
					i++;
				}
				else 
					result.append(lines[i]).append(END_LINE);
			}
			else {
				result.append(lines[i]).append(END_LINE);
				temp.add(line);
			}
		}
		return result.toString();
	}

	private void writeToFile(String output, String fileName) {
		try {
			File file = new File(this.srcPath+"/testCases");
			if (!file.exists()) {
				if (file.mkdirs()) {
					System.out.println("Directory is created!");
				} else {
					System.out.println("Failed to create directory!");
				}
			}

			FileWriter fstream = new FileWriter(srcPath+"/testCases/"+fileName);
			BufferedWriter buffOut = new BufferedWriter(fstream);

			System.out.println("Generating...");
			System.out.println();


			buffOut.append(output);					
			buffOut.close();

		} catch (Exception e) {
			e.printStackTrace();
		}	
	}

	private boolean enableFiltering(String packageName) {
		if (srcPath.endsWith("AnyCut") || srcPath.endsWith("TippyTipper") 
				|| packageName.endsWith("mandaria") || packageName.endsWith("anycut")
				|| packageName.endsWith("remembeer") || srcPath.endsWith("Remembeer")
				|| packageName.endsWith("munchlife") || srcPath.endsWith("MunchLife") || packageName.endsWith("httpmon")) {
			return true;
		}
		return false;
	}

	private String extractTitle(String stringID) {
		// open strings.xml
		Document stringsXmlDoc = prepareXML(srcPath + "/res/values/strings.xml");
		NodeList strings = stringsXmlDoc.getElementsByTagName("string");

		if(strings != null && strings.getLength() > 0) {
			for(int k = 0 ; k < strings.getLength(); k++) {
				Element element = (Element)strings.item(k);

				String stingName = element.getAttribute("name");
				if(stringID.equals(stingName)) {
					return element.getTextContent();
				}

			}
		} else {
			System.out.println("THIS SHOULD NOT HAPPEN: EMPTY ARRAY");
		}
		return "";
	}

    private String extractRFilePackage(String targetAppPath) throws IOException, SAXException, ParserConfigurationException {
        return Util.extractAppPackage(targetAppPath);
    }
}
